from . import AbstractComparer, Status


class DummyComparer(AbstractComparer):
    """A comparer that always returns [`Status.CHANGED`][web_watchr.compare.Status]."""

    def __call__(self, text: str) -> Status:
        """Always return [`Status.CHANGED`][web_watchr.compare.Status]."""
        return Status.CHANGED



from .abstract_comparer import AbstractComparer, Status
from .dummy_comparer import DummyComparer
from .fs_comparer import FSComparer

__all__ = ["AbstractComparer", "DummyComparer", "FSComparer", "Status"]



from abc import ABC, abstractmethod

from pydantic import BaseModel


class AbstractAlerter(BaseModel, ABC):
    """Send an alert."""

    @abstractmethod
    def __call__(self, text: str) -> None:
        """Send an alert.

        A new alerter must implement this method. The method should send the alert with the given text.

        Args:
            text: The text of the alert.
        """
        ...



from pathlib import Path

import pytest
from web_watchr.compare.fs_comparer import FSComparer, Status


@pytest.mark.parametrize(
    "exists",
    [True, False],
)
def test_cache_path(
    exists: bool,
    fs_comparer: FSComparer,
    cache_dir: Path,
):
    assert not cache_dir.exists()
    if exists:
        cache_dir.mkdir(parents=True)

    cache_path = fs_comparer.cache_path

    assert cache_dir.exists()
    assert cache_path.parent == cache_dir
    assert not cache_path.exists()


@pytest.mark.parametrize(
    "send_on_missing, status",
    [
        (True, Status.CHANGED),
        (False, Status.NO_CHANGES),
    ],
)
def test_first_call(
    send_on_missing: bool,
    status: Status,
    fs_comparer: FSComparer,
):
    text = "anything"
    fs_comparer.send_on_missing = send_on_missing

    received_status = fs_comparer(text=text)

    assert received_status == status


@pytest.mark.parametrize(
    "texts, status",
    [
        (["a", "b", "c"], Status.CHANGED),
        (["a", "a", "b"], Status.CHANGED),
        (["a", "a", "a"], Status.NO_CHANGES),
    ],
)
def test_second_call(
    texts: list[str],
    status: Status,
    fs_comparer: FSComparer,
):
    for text in texts:
        received_status = fs_comparer(text=text)

    assert received_status == status



from unittest.mock import Mock, patch

from web_watchr.alert.telegram_alerter import TelegramAlerter


@patch("web_watchr.alert.telegram_alerter.Bot")
@patch("web_watchr.alert.telegram_alerter.run")
def test_telegram_alerter(
    mock_run: Mock,
    mock_telegram_bot: Mock,
):
    token = "invalid"
    chat_id = "different_but_invalid"
    message = "test message"
    alerter = TelegramAlerter(token=token, chat_id=chat_id)

    alerter(message)

    mock_telegram_bot.assert_called_once_with(token=token)
    mock_telegram_bot.return_value.send_message.assert_called_once_with(
        chat_id=chat_id,
        text=message,
    )
    mock_run.assert_called_once_with(
        mock_telegram_bot.return_value.send_message.return_value
    )



from abc import ABC, abstractmethod
from enum import Enum, auto

from pydantic import BaseModel


class Status(Enum):
    """Enum representing the status of a comparison.

    Convenience enum because I got confused with booleans.

    Attributes:
        CHANGED: The text has changed.
        NO_CHANGES: The text has not changed.
    """

    CHANGED = auto()
    NO_CHANGES = auto()


class AbstractComparer(BaseModel, ABC):
    """Decide if a text has changed."""

    @abstractmethod
    def __call__(self, text: str) -> Status:
        """Decide if a text has changed.

        A new comparer must implement this method. The method should compare the text with the previous
        state and return the status of the comparison.
        """
        ...



from . import AbstractAlerter


class PrintAlerter(AbstractAlerter):
    """Print the alert to the standard output."""

    def __call__(self, text: str) -> None:
        """Print the alert to the standard output.

        Args:
            text: The text of the alert.
        """
        print(text)



from typing import Callable

from loguru import logger
from playwright.sync_api import Playwright, sync_playwright
from pydantic import BaseModel, Field, PrivateAttr

from web_watchr.alert import AbstractAlerter, PrintAlerter
from web_watchr.compare import AbstractComparer, FSComparer, Status


class Watchr(BaseModel):
    """Central class that orchestrates the polling, comparing, and alerting process.

    Attributes:
        alerter: The instance is called with the scraped text and is responsible to send out the alert.
        comparer: The instance is called with the scraped text and is responsible to compare the text with
            the previous state to check if there are any changes and an alert is necessary.
    """

    comparer: AbstractComparer = Field(default_factory=FSComparer)
    alerter: AbstractAlerter = Field(default_factory=PrintAlerter)

    _poller: Callable[[Playwright], str] | None = PrivateAttr(default=None)

    @property
    def poller(self) -> Callable[[Playwright], str]:
        """The poller function that scrapes the text from the website."""
        if self._poller is None:
            message = "Please set a poller function using the `set_poller` decorator."
            logger.error(message)
            raise ValueError(message)
        return self._poller

    def set_poller(
        self,
        f: Callable[[Playwright], str],
    ) -> Callable[[Playwright], str]:
        """Decorator to set the poller function.

        The wrapper returns the function unchanged, but stores it in the `_poller` attribute.

        Args:
            f: A function that takes a `Playwright` instance and returns the scraped text.

        Returns:
            The function that was passed in.
        """
        self._poller = f
        return f

    def __call__(
        self,
    ) -> None:
        """Main method that orchestrates the polling, comparing, and alerting process.

        When called the poller function is called to scrape the text from the website. The text is then
        compared to the previous state. If there are changes, an alert is sent out.
        """
        with sync_playwright() as playwright:
            result = self.poller(playwright)
            logger.debug(result)

        if self.comparer(result) == Status.NO_CHANGES:
            logger.info("No changes detected.")
            return

        logger.info("Changes detected! Sending alert.")
        self.alerter(result)



import string
from random import choices, randint

from web_watchr.compare import DummyComparer, Status


def test_always_status_changed():
    comparer = DummyComparer()
    alphabet = string.ascii_letters + string.digits
    text_length = randint(0, 10)
    text = "".join(choices(alphabet, k=text_length))

    assert comparer(text=text) == Status.CHANGED



import builtins
from unittest.mock import Mock

from web_watchr.alert.print_alerter import PrintAlerter


def test_std_out_alerter():
    message = "Test"
    alerter = PrintAlerter()

    # Inspired by https://stackoverflow.com/a/62360735/9685500
    mock = Mock()
    print_original = print
    builtins.print = mock

    try:
        alerter(message)

        mock.assert_called_once_with(message)
    finally:
        builtins.print = print_original  # ensure print is "unmocked"



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterator
from unittest.mock import MagicMock

import pytest
from web_watchr.alert.abstract_alerter import AbstractAlerter
from web_watchr.compare.abstract_comparer import AbstractComparer, Status
from web_watchr.compare.fs_comparer import FSComparer
from web_watchr.watchr import Watchr


@pytest.fixture()
def cache_dir() -> Iterator[Path]:
    with TemporaryDirectory() as temp_dir:
        yield Path(temp_dir) / "cache"


@pytest.fixture()
def fs_comparer(
    cache_dir: Path,
) -> FSComparer:
    return FSComparer(cache_dir=cache_dir)


@pytest.fixture()
def mock_variable_poller() -> MagicMock:
    poller = MagicMock()
    poller.side_effect = ["a", "b", "c"]
    return poller


@pytest.fixture()
def mock_status_list() -> list[Status]:
    return [Status.CHANGED, Status.NO_CHANGES, Status.CHANGED]


@pytest.fixture()
def mock_comparer(
    mock_status_list: list[Status],
) -> MagicMock:
    mock = MagicMock(spec=AbstractComparer)
    mock.side_effect = mock_status_list
    return mock


@pytest.fixture()
def mock_alerter() -> MagicMock:
    return MagicMock(spec=AbstractAlerter)


@pytest.fixture()
def mocked_watchr(
    mock_comparer: MagicMock,
    mock_alerter: MagicMock,
) -> Watchr:
    return Watchr(
        comparer=mock_comparer,
        alerter=mock_alerter,
    )



from unittest.mock import MagicMock

import pytest
from web_watchr.compare.abstract_comparer import Status
from web_watchr.watchr import Watchr


def test_unset_poller(
    mocked_watchr: Watchr,
):
    with pytest.raises(ValueError):
        mocked_watchr()


def test_watchr(
    mocked_watchr: Watchr,
    mock_variable_poller: MagicMock,
    mock_status_list: list[Status],
):
    mocked_watchr.set_poller(mock_variable_poller)
    n = len(mock_status_list)
    n_sent = len([status for status in mock_status_list if status == Status.CHANGED])

    for _ in range(n):
        mocked_watchr()

    assert mock_variable_poller.call_count == n
    assert mocked_watchr.comparer.call_count == n
    assert mocked_watchr.alerter.call_count == n_sent



import os

from playwright.sync_api import Playwright
from web_watchr import Watchr
from web_watchr.alert import TelegramAlerter

watchr = Watchr(
    alerter=TelegramAlerter(
        token=os.getenv("TELEGRAM_TOKEN"),
        chat_id=os.getenv("TELEGRAM_CHAT_ID"),
    )
)


@watchr.set_poller
def poll(playwright: Playwright) -> str:
    browser = playwright.chromium.launch(headless=True)
    context = browser.new_context()
    page = context.new_page()
    page.goto("https://www.example.com/")
    text = page.get_by_role("heading").inner_text()
    context.close()
    browser.close()

    return text


if __name__ == "__main__":
    watchr()



from pathlib import Path

from loguru import logger

from web_watchr.compare.abstract_comparer import Status

from . import AbstractComparer


class FSComparer(AbstractComparer):
    """A comparer that compares the text with the content of a file.

    Attributes:
        cache_dir: The directory where the cache file is stored. Defaults to `~/.local/share/web_watchr/cache`.
        send_on_missing: Whether to send a notification if the cache file is missing.
        identifier: The identifier used for the cache file.
    """

    cache_dir: Path = Path("~/.local/share/web_watchr/cache").expanduser()
    send_on_missing: bool = False
    identifier: str = "fs_comparer"

    _cache_path: Path | None = None

    @property
    def cache_path(self) -> Path:
        """The path to the cache file.

        Creates the cache directory if it does not exist.
        """
        if self._cache_path is None:
            self._cache_path = self.cache_dir / f"{self.identifier}.txt"
            if not self.cache_dir.exists():
                logger.debug(f"Creating cache directory: {self.cache_dir}")
                self.cache_dir.mkdir(parents=True)

        return self._cache_path

    def __call__(self, text: str) -> Status:
        """Compare the text with the content of the cache file.

        Args:
            text: The text to compare with the cache file.

        Returns:
            [`Status.CHANGED`][web_watchr.compare.Status] if the text is different from the cache file.
            [`Status.NO_CHANGES`][web_watchr.compare.Status] if the text is the same as the cache file.
            If the cache file is missing, it returns [`Status.CHANGED`][web_watchr.compare.Status] if `send_on_missing` is `True`.
        """
        if not self.cache_path.exists():
            logger.debug(f"Cache file not found: {self.cache_path}. Saving new state.")
            self._save(new_state=text)
            return Status.CHANGED if self.send_on_missing else Status.NO_CHANGES

        if not self._equal_to_cache(new_state=text):
            self._save(new_state=text)
            return Status.CHANGED

        return Status.NO_CHANGES

    def _load(self) -> str:
        with open(self.cache_path, "r") as f:
            return f.read()

    def _equal_to_cache(self, *, new_state: str) -> bool:
        old_state = self._load()
        return old_state == new_state

    def _save(self, *, new_state: str):
        with open(self.cache_path, "w") as f:
            f.write(new_state)



from asyncio import run

from loguru import logger
from telegram import Bot

from . import AbstractAlerter


class TelegramAlerter(AbstractAlerter):
    """Sends a message to a telegram chat.

    For this alerter to work, you need to
    [create a telegram bot and get its token](https://core.telegram.org/bots/tutorial#obtain-your-bot-token).
    Furthermore, the bot needs to be added to the chat you want to send messages to and you
    need to retrieve the chat ID of the chat (e.g., like [this](https://stackoverflow.com/a/32572159/9685500)).

    Attributes:
        token: The token of the telegram bot.
        chat_id: The chat ID to send the message to.
    """

    token: str
    chat_id: str

    _bot: Bot | None = None

    @property
    def bot(self) -> Bot:
        """Telegram bot instance"""
        if self._bot is None:
            self._bot = Bot(token=self.token)
        return self._bot

    def __call__(self, text: str) -> None:
        """Send a message to a telegram chat.

        Args:
            text: The text of the message.
        """
        logger.debug(f"Sending message '{text}' to telegram.")
        run(self.bot.send_message(chat_id=self.chat_id, text=text))
